from __future__ import annotations

import copy
import json
import logging
import os
import shv
import typing

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    model_serializer,
    NonNegativeInt,
)

from g3elements import Zone, Route, Gate

from ..utils import SitesFileDownloader, SitesFilePaths
from ._list_operations import ListOperations


logger = logging.getLogger('typeInfo')


class TypeInfoModel(BaseModel):
    """Root model for TypeInfo models."""
    model_config = ConfigDict(
        extra='allow',
        arbitrary_types_allowed=True,
        validate_assignment=True
        )

    meta: typing.Mapping[str | int, typing.Any] = Field(
        default={}, exclude=True
        )

    @staticmethod
    def _unpack_cpon(data: typing.Any) -> shv.SHVMapType:
        converted = shv.cpon.Cpon.unpack(data)
        if shv.is_shvmap(converted):
            return converted
        raise TypeError('Invalid TypeInfo model structure (expected SHVMap)')

    @classmethod
    def model_validate_cpon(
        cls,
        cpon_data: str | bytes | bytearray,
        *,
        strict: bool | None = None,
        from_attributes: bool | None = None,
        context: dict[str, typing.Any] | None = None
    ) -> typing.Self:
        unpacked = cls._unpack_cpon(cpon_data)
        model = cls.model_validate(
            unpacked,
            strict=strict,
            from_attributes=from_attributes,
            context=context
            )
        if hasattr(unpacked, '_meta'):
            model.meta = unpacked._meta
        return model

    @model_serializer(when_used='json-unless-none')
    def _serialize_shvtype(self) -> typing.Any:
        values: dict[str, typing.Any] = {}
        for field in self.model_fields_set:
            value = getattr(self, field)
            if shv.is_shvbool(value):
                values[field] = bool(value)
            elif shv.is_shvnull(value):
                values[field] = None
            else:
                values[field] = value
        return values

    def add_attrs(self, **new_attrs) -> typing.Self:
        """Add new attributes to the model instance and return an updated copy.

        Returns:
            typing.Self: Updated model instance.
        """
        attrs = self.model_dump(exclude_unset=True)
        attrs.update(new_attrs)
        return self.__class__(**attrs)

    def model_dump_cpon(
        self,
        *,
        indent: NonNegativeInt | None = None,
        indent_char: str = '\t',
        include: (
            set[int] |
            set[str] |
            dict[int, typing.Any] |
            dict[str, typing.Any] |
            None
        ) = None,
        exclude: (
            set[int] |
            set[str] |
            dict[int, typing.Any] |
            dict[str, typing.Any] |
            None
        ) = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        round_trip: bool = False,
        warnings: bool = True
    ) -> str:
        data = self.model_dump(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            round_trip=round_trip,
            warnings=warnings
            )
        indentation = indent_char * indent if indent is not None else ''
        cpon_options = shv.CponWriter.Options(indent=indentation.encode())
        contents = shv.SHVMeta.new(data, meta=self.meta)  # type: ignore
        return shv.cpon.Cpon.pack(contents, cpon_options)


# Device descriptions

class TypeInfoDeviceDescriptionPropertyMethod(TypeInfoModel):
    accessGrant: str | None = None
    description: str | None = None
    flags: NonNegativeInt | None = None
    label: str | None = None
    name: str | None = None
    signature: int | None = None


class TypeInfoDeviceDescriptionProperty(TypeInfoModel):
    alarm: str | None = None
    autoload: shv.SHVBool | bool | None = None
    description: str | None = None
    label: str | None = None
    methods: list[TypeInfoDeviceDescriptionPropertyMethod] | None = None
    monitorOptions: dict[str, typing.Any] | None = None
    monitored: shv.SHVBool | bool | None = None
    name: str | None = None
    typeName: str | None = None
    unit: str | None = None


class TypeInfoDeviceDescription(TypeInfoModel):
    properties: list[TypeInfoDeviceDescriptionProperty] | None = None

    def get_prop(
        self, prop_name: str
    ) -> TypeInfoDeviceDescriptionProperty:
        if self.properties is None:
            raise ValueError("properties is None")
        return ListOperations(self.properties).get(item_name=prop_name)

    @property
    def root_prop(self) -> TypeInfoDeviceDescriptionProperty:
        if self.properties is None:
            raise ValueError("properties is None")
        return ListOperations(self.properties).get(item_name='')

    @property
    def status_prop(self) -> TypeInfoDeviceDescriptionProperty:
        if self.properties is None:
            raise ValueError("properties is None")
        return ListOperations(self.properties).get(item_name='status')

    @status_prop.setter
    def status_prop(
        self, new_status_prop: TypeInfoDeviceDescriptionProperty
    ) -> None:
        if self.properties is None:
            raise ValueError("properties is None")
        ListOperations(self.properties).add(
            item_name="status",
            item_data=new_status_prop,
            replace_if_exists=True
            )

    def get_root_method(
        self, meth_name: str
    ) -> TypeInfoDeviceDescriptionPropertyMethod:
        if self.root_prop.methods is None:
            raise ValueError("root_prop.methods is None")
        return ListOperations(self.root_prop.methods).get(item_name=meth_name)

    def del_prop(self, prop_name: str) -> None:
        if self.properties is None:
            raise ValueError("properties is None")
        ListOperations(self.properties).delete(item_name=prop_name)

    def del_root_method(self, meth_name: str) -> None:
        if self.root_prop.methods is None:
            raise ValueError("root_prop.methods is None")
        ListOperations(self.root_prop.methods).delete(item_name=meth_name)


# Types

class TypeInfoTypeField(TypeInfoModel):
    alarm: str | None = None
    alarmLevel: int | None = None
    description: str | None = None
    label: str | None = None
    name: str | None = None
    typeName: str | None = None
    value: typing.Any | None = None


class TypeInfoType(TypeInfoModel):
    fields: list[TypeInfoTypeField] | None = None
    name: str | None = None
    typeName: str | None = None

    def get_field(self, field_name: str) -> TypeInfoTypeField:
        if self.fields is None:
            raise ValueError("No fields to get")
        return ListOperations(self.fields).get(item_name=field_name)

    def del_field(self, field_name: str) -> None:
        if self.fields is None:
            raise ValueError("No fields to get")
        ListOperations(self.fields).delete(item_name=field_name)


# TypeInfo (main)

class TypeInfo(TypeInfoModel):
    """'TypeInfo.cpon' file data model."""
    blacklistedPaths: dict[str, None]
    deviceDescriptions: dict[str, TypeInfoDeviceDescription]
    devicePaths: dict[str, str]
    extraTags: dict[str, dict[str, str]]
    systemPathsRoots: dict[str, typing.Any]
    types: dict[str, TypeInfoType]

    @classmethod
    def model_validate_local_file(
        cls,
        filepath: str,
        *,
        strict: bool | None = None,
        context: dict[str, typing.Any] | None = None
    ) -> typing.Self:
        with open(filepath, 'r', encoding='utf-8') as file:
            data = file.read()
        return cls.model_validate_cpon(data, strict=strict, context=context)

    @classmethod
    def model_validate_sites_repo(
        cls,
        branch: str = 'master',
        *,
        strict: bool | None = None,
        from_attributes: bool | None = None,
        context: dict[str, typing.Any] | None = None
    ) -> typing.Self:
        data = SitesFileDownloader().read(SitesFilePaths.TYPE_INFO, branch)
        assert isinstance(data, str)
        return cls.model_validate_cpon(
            data,
            strict=strict,
            from_attributes=from_attributes,
            context=context
            )

    @staticmethod
    def is_decimal_type(type_data: TypeInfoType) -> bool:
        return hasattr(type_data, 'decPlaces')

    @staticmethod
    def is_enum_type(type_data: TypeInfoType) -> bool:
        return getattr(type_data, 'typeName', '').casefold() == 'enum'

    @staticmethod
    def is_bitfield_type(type_data: TypeInfoType) -> bool:
        return getattr(type_data, 'typeName', '').casefold() == 'bitfield'

    def get_status_type_from_device_name(self, name: str) -> TypeInfoType:
        device_desc = self.deviceDescriptions[name]
        status_name = device_desc.status_prop.typeName
        assert status_name is not None
        return self.types[status_name]

    @staticmethod
    def _update_device_description(
        device_description: TypeInfoDeviceDescription,
        exclude_root_methods: list[str] | None = None,
        exclude_properties: list[str] | None = None,
    ) -> None:
        if exclude_root_methods:
            for meth in exclude_root_methods:
                device_description.del_root_method(meth)
        if exclude_properties:
            for prop in exclude_properties:
                device_description.del_prop(prop)

    @staticmethod
    def _update_status_type(
        status_type: TypeInfoType,
        exclude_status_fields: list[str]
    ) -> None:
        for field in exclude_status_fields:
            status_type.del_field(field)

    def update_device(
        self,
        device_name: str,
        exclude_root_methods: list[str] | None = None,
        exclude_properties: list[str] | None = None,
        exclude_status_fields: list[str] | None = None
    ) -> None:
        description = self.deviceDescriptions[device_name]
        self._update_device_description(
            description, exclude_root_methods, exclude_properties
            )
        if exclude_status_fields:
            status_type = self.get_status_type_from_device_name(device_name)
            self._update_status_type(status_type, exclude_status_fields)

    def create_child_device(
        self,
        device_name: str,
        parent_device_name: str,
        exclude_root_methods: list[str] | None = None,
        exclude_properties: list[str] | None = None,
        exclude_status_fields: list[str] | None = None
    ) -> None:
        # create and update the child device description
        # as a copy of the parent device description
        parent_description = self.deviceDescriptions[parent_device_name]
        child_description = copy.deepcopy(parent_description)
        self._update_device_description(
            child_description, exclude_root_methods, exclude_properties
            )
        # add reference to the parent device description
        child_description = child_description.add_attrs(
            restrictionOfDevice=parent_device_name
            )
        # add the new child device description to device descriptions
        self.deviceDescriptions[device_name] = child_description
        if exclude_status_fields:
            # create and update the child status type
            # as a copy of the parent status type
            parent_status_type = self.get_status_type_from_device_name(
                parent_device_name
                )
            child_status_type = copy.deepcopy(parent_status_type)
            self._update_status_type(child_status_type, exclude_status_fields)
            # udpate status type name in the child description
            parent_status_name = child_description.status_prop.typeName
            child_status_name = f'Status{device_name}'
            child_description.status_prop.typeName = child_status_name
            # add reference to the parent status type
            child_status_type = child_status_type.add_attrs(
                restrictionOfType=parent_status_name
                )
            # add the new status type to TypeInfo types
            self.types[child_status_name] = child_status_type

    def update_superior_system(
        self,
        exclude_methods: list[str] | None = None,
        exclude_status_fields: list[str] | None = None
    ) -> None:
        system = self.deviceDescriptions['System_G3']
        if exclude_methods:
            sup_system_prop = system.get_prop('SuperiorSystem')
            assert sup_system_prop.methods is not None
            list_ops_methods = ListOperations(sup_system_prop.methods)
            for meth in exclude_methods:
                list_ops_methods.delete(meth)
        if exclude_status_fields:
            sup_system_status = self.types['StatusSuperiorSystem']
            self._update_status_type(sup_system_status, exclude_status_fields)

    def update_from_system_config(
        self, system_config: dict, zone_name: str
    ) -> None:
        logger.info(
            'Updating TypeInfo CPON file from system config data (zone "%s")',
            zone_name
            )
        SystemDictConnector(zone_name, system_config, self).update()

    def update_from_system_config_local_file(
        self, file_path: str, zone_name: str
    ) -> None:
        logger.info(
            'Parsing system config file "%s"', os.path.abspath(file_path)
            )
        with open(file_path, 'r') as file:
            system_config = json.load(file)
        self.update_from_system_config(system_config, zone_name)


class SystemDictConnector:
    def __init__(
        self,
        zone_name: str,
        system_data: dict,  # SystemDict
        type_info: TypeInfo
    ) -> None:
        self.system = system_data
        self.zone = Zone.from_system_config(zone_name, system_data)  # type: ignore  # noqa
        self.type_info = type_info

    def update_device_descriptions(self) -> None:
        logger.info("Updating deviceDescriptions...")
        for type_data in self.system['Visu'].values():
            exclude_root_methods = []
            exclude_properties = []
            exclude_status_fields = []
            for var_data in type_data['vars'].values():
                var = var_data['name']
                if not var_data['is_deleted']:
                    continue
                if var.startswith('method'):
                    exclude_root_methods.append(var.removeprefix('method/'))
                elif var.startswith('status'):
                    exclude_status_fields.append(var.removeprefix('status/'))
                else:
                    exclude_properties.append(var)
            if type_data['restricted_to']:
                self.type_info.create_child_device(
                    device_name=type_data['type'],
                    parent_device_name=type_data['restricted_to'],
                    exclude_properties=exclude_properties,
                    exclude_root_methods=exclude_root_methods,
                    exclude_status_fields=exclude_status_fields
                    )
            else:
                self.type_info.update_device(
                    device_name=type_data['type'],
                    exclude_properties=exclude_properties,
                    exclude_root_methods=exclude_root_methods,
                    exclude_status_fields=exclude_status_fields
                )
        logger.info("deviceDescriptions updated successfully.")

    def update_device_paths(self) -> None:
        logger.info("Updating device paths...")
        self.type_info.devicePaths.clear()
        for element in self.zone.all_elements:
            path = element.shv_path
            visu_data = element.data.get('device_dict', {}).get('visu', {})
            visu_type = visu_data.get('shvDeviceType', '')
            if path and (visu_type or 'system' in path.lower()):
                if(path == "systemSafety"):
                    path_systemSafety = "systemSafety/SafeLogic"
                    self.type_info.devicePaths[path_systemSafety] = visu_type
                else:
                    self.type_info.devicePaths[path] = visu_type
        logger.info("Device paths updated successfully.")

    def update_gpio(self) -> None:
        logger.info("Updating GPIO device types...")
        gpio_inputs = self.type_info.types['GPIOInputs'].add_attrs(
            siteSpecificLocalization=True
            )
        self.type_info.types['GPIOInputs'] = gpio_inputs
        gpio_outputs = self.type_info.types['GPIOOutputs'].add_attrs(
            siteSpecificLocalization=True
            )
        self.type_info.types['GPIOOutputs'] = gpio_outputs
        logger.info("GPIO device types updated successfully.")

    def update_status_buttons(self) -> None:
        logger.info("Updating StatusButtons device type...")
        buttons = self.type_info.types['StatusButtons'].add_attrs(
            siteSpecificLocalization=True
            )
        self.type_info.types['StatusButtons'] = buttons
        logger.info("StatusButtons device updated successfully.")

    def update_extra_tags(self) -> None:
        logger.info("Updating extraTags...")
        tags = self.type_info.extraTags
        tags.clear()
        if self.zone.is_safety:
            key = f'{self.zone.shv_path}/method/setNormal'
            tags[key] = {"safetyManager": "systemSafety"}
        for route in self.zone.find_by_type(Route):
            tags[route.shv_path] = {"startGate": route.layout.entry_gate.name}
        logger.info("extraTags updated successfully.")

    def update_blacklisted_paths(self) -> None:
        logger.info("Updating blacklistedPaths...")
        # remove default
        default = 'devices/zone/Zone/gate/G01/requestMemory'
        if default in self.type_info.blacklistedPaths:
            del self.type_info.blacklistedPaths[default]
        # add actual gates
        for gate in self.zone.find_by_type(Gate):
            gate_visu_data = gate.data.get('device_dict', {}).get('visu', {})
            gate_visu_type_name = gate_visu_data.get('shvDeviceType', '')
            if not gate_visu_type_name:
                logger.warning(f'Gate "{gate.name}" does not have a Visu type')
                continue
            gate_visu_type = self.system['Visu'][gate_visu_type_name]
            request_memory = gate_visu_type['vars']['requestMemory']
            if request_memory['is_blacklisted']:
                key = f'{gate.shv_path}/requestMemory'
                self.type_info.blacklistedPaths[key] = None
        logger.info("blacklistedPaths updated successfully.")

    def update(self) -> None:
        logger.info("Updating typeInfo...")
        self.update_device_descriptions()
        self.update_device_paths()
        self.update_gpio()
        self.update_status_buttons()
        self.update_extra_tags()
        self.update_blacklisted_paths()
        logger.info("typeInfo updated successfully.")



from .typeinfo import TypeInfo
from .site import SiteSVG
from .meta import MetaZone, MetaProject
from . import meta, typeinfo, site, utils


__all__ = [
    'TypeInfo',
    'SiteSVG',
    'MetaZone',
    'MetaProject',
    'meta',
    'typeinfo',
    'site',
    'utils'
]



import logging
import typing

from abc import ABC, abstractmethod
from pydantic import BaseModel, ConfigDict, Field
from pydantic.alias_generators import to_camel


logger = logging.getLogger('g3visu.meta')


class BaseMeta(BaseModel, ABC):
    """Base class for meta data classes."""
    model_config = ConfigDict(
        alias_generator=to_camel,
        validate_assignment=True,
        populate_by_name=True,
        )

    @classmethod
    @abstractmethod
    def model_validate_system_config(cls, system_config: dict) -> typing.Self:
        pass


class MetaProjectSiteWare(BaseMeta):
    shv_user: str = Field(default="")
    shv_password: str = Field(default="")
    default_language: str
    language_array: list[str]

    @classmethod
    def model_validate_system_config(
        cls, system_config: dict, **kwargs
    ) -> typing.Self:
        project_dict: dict = system_config['Software']['Common']['Project']
        project_data: dict = list(project_dict.values())[0]
        language = project_data['general']['language']
        language_array = ['en', language]

        if str(kwargs['shv_user'])=='None':
            shv_password = ""
            shv_user = ""
            logger.info("No default shv user name defined, using empty string")
        else:
            shv_user = kwargs['shv_user']
            if str(kwargs['shv_password'])=='None':
                shv_password = ""
                logger.info("No default shv user password defined, using empty string")
            else:
                shv_password = kwargs['shv_password']
        return cls(shv_user=shv_user, shv_password=shv_password, default_language=language, language_array=language_array)


class MetaProject(BaseMeta):
    """'_meta.json' file data model for the project."""
    name: str
    siteware: MetaProjectSiteWare

    @classmethod
    def model_validate_system_config(
        cls, system_config: dict, **kwargs
    ) -> typing.Self:
        project_dict: dict = system_config['Software']['Common']['Project']
        project_data: dict = list(project_dict.values())[0]
        name = project_data['general']['location']
        if not name:
            name = project_data['general']['name']
        siteware = MetaProjectSiteWare.model_validate_system_config(
            system_config, shv_user=kwargs['shv_user'], shv_password=kwargs['shv_password']
            )
        return cls(name=name, siteware=siteware)


class MetaZone(BaseMeta):
    """'_meta.json' file data model for the zone."""
    type_: str = Field(default="DepotG3", alias="type")
    name: str = Field(default="")
    gps: tuple[float, float] = Field(default=(0.0, 0.0))
    hp3: dict[str, str] = Field(default={"syncPath": ".app/history"}, alias="HP3")
    visu: dict[str, str] = Field(default={"svgPath": "_files/site.svg"})

    @classmethod
    def model_validate_system_config(
        cls, system_config: dict, **kwargs
    ) -> typing.Self:

        if 'name' not in kwargs:
            logger.warning("No zone name provided, using empty string")
            return cls(gps=gps)
        zone_name = kwargs['name']


        system_dict: dict = system_config['Software']['Common']['System']
        system_data: dict = list(system_dict.values())[0]

        gps_location = system_data.get('general', {}).get('gps_location')

        if gps_location is None:
            system_dict: dict = system_config['Software'][zone_name]['Zone']
            system_data: dict = list(system_dict.values())[0]

        try:
            gps_str = system_data['general']['gps_location']
            gps_str_split = gps_str.split(',')
            gps = (float(gps_str_split[0]), float(gps_str_split[1]))
        except Exception as e:
            logger.warning(f"Could not parse gps from system config: {e}")
            gps = (0.0, 0.0)

        if zone_name not in system_config['Software']:
            logger.error(
                f"Zone name not found in system config: {zone_name}. "
                f"Using empty string"
                )
            return cls(gps=gps)
        zone_dict = system_config['Software'][zone_name]['Zone']
        zone_data = list(zone_dict.values())[0]
        visu_name = zone_data.get('visu', {}).get('name', '')
        if not visu_name:
            visu_name = zone_data['general']['name']
        return cls(name=visu_name, gps=gps)



from ._typeinfo import (
    TypeInfoDeviceDescription,
    TypeInfoDeviceDescriptionProperty,
    TypeInfoDeviceDescriptionPropertyMethod,
    TypeInfoType,
    TypeInfoTypeField,
    TypeInfo
)


__all__ = [
    'TypeInfoDeviceDescription',
    'TypeInfoDeviceDescriptionProperty',
    'TypeInfoDeviceDescriptionPropertyMethod',
    'TypeInfoType',
    'TypeInfoTypeField',
    'TypeInfo'
]



import typing

from pydantic import BaseModel


T = typing.TypeVar('T', bound=BaseModel)


class ListOperations(typing.Generic[T]):
    def __init__(self, data: list[T]) -> None:
        self.data: list[T] = data

    def get(self, item_name: str) -> T:
        for item in self.data:
            if not isinstance(item, BaseModel):
                continue
            if getattr(item, 'name', None) == item_name:
                return item
        raise KeyError(item_name)

    def add(
        self,
        item_name: str,
        item_data: T,
        raise_if_exists: bool = False,
        replace_if_exists: bool = False
    ) -> None:
        try:  # check if the item is already present in the list
            self.get(item_name)
            if raise_if_exists:
                raise ValueError(f'Item "{item_name}" already exists.')
            if replace_if_exists:
                self.delete(item_name)
        except KeyError:
            pass
        self.data.append(item_data)

    def delete(self, item_name: str) -> None:
        for item in self.data:
            if not isinstance(item, BaseModel):
                continue
            if getattr(item, 'name', None) == item_name:
                self.data.remove(item)
                break



from ._download_from_sites import (
    AccessToken,
    SitesFileDownloader,
    SitesFilePaths,
    ensure_token_set,
    prompt_token_until_valid,
)


__all__ = [
    "AccessToken",
    "SitesFileDownloader",
    "SitesFilePaths",
    "ensure_token_set",
    "prompt_token_until_valid",
]



import argparse
import json
import os
import logging

from .meta import MetaProject, MetaZone
from .site import SiteSVG
from .typeinfo import TypeInfo
from .utils import ensure_token_set


logger = logging.getLogger('g3visu.meta')


def create_meta_files(
    system_config: dict, zone_name: str, output_dir: str, shv_user: str, shv_password: str
) -> None:
    # create the meta project and meta zone file data
    meta_project = MetaProject.model_validate_system_config(
        system_config, shv_user=shv_user, shv_password=shv_password
        )
    meta_zone = MetaZone.model_validate_system_config(
        system_config, name=zone_name
        )
    # create output directories and file paths
    meta_project_path = os.path.join(output_dir, '_meta.json')
    zone_dir = os.path.join(output_dir, zone_name)
    meta_zone_path = os.path.join(zone_dir, '_meta.json')
    os.makedirs(zone_dir, exist_ok=True)
    # write the file data to the output files
    with open(meta_project_path, 'w', encoding='utf-8') as f:
        f.write(meta_project.model_dump_json(indent=4, by_alias=True))
    with open(meta_zone_path, 'w', encoding='utf-8') as f:
        f.write(meta_zone.model_dump_json(indent=4, by_alias=True))


def create_site_svg_file(
    system_config: dict,
    zone_name: str,
    output_dir: str,
    visu_types_path: str | None = None
) -> None:
    # create the site svg file and update it from the system config
    if visu_types_path:
        site_svg = SiteSVG.from_local(visu_types_path)
    else:
        site_svg = SiteSVG.from_sites_repo()
    site_svg.update_from_system_config(system_config, zone_name)  # type: ignore  # noqa
    # write the file data to the output file
    site_svg.to_file(file_path=os.path.join(output_dir, 'site.svg'))


def create_type_info_file(
    system_config: dict, zone_name: str, output_dir: str
) -> None:
    # create the type info file and update it from the system config
    type_info = TypeInfo.model_validate_sites_repo()
    type_info.update_from_system_config(system_config, zone_name)  # type: ignore  # noqa
    # write the file data to the output file
    type_info_path = os.path.join(output_dir, 'typeInfo.cpon')
    with open(type_info_path, 'w', encoding='utf-8') as f:
        f.write(type_info.model_dump_cpon(indent=1, exclude_unset=True))


def valid_system_config(system_config_arg: str) -> tuple[str, str]:
    parts = system_config_arg.strip().split('@')
    if len(parts) != 2:
        raise argparse.ArgumentTypeError(
            f'System config path argument must be formatted as '
            f'"<path>@<zone_name>", got "{system_config_arg}".'
            )
    if not all(parts):
        raise argparse.ArgumentTypeError(
            f'Each part separated by "@" must be non-empty in the argument '
            f'"{system_config_arg}".'
            )
    return (parts[0], parts[1])


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Generate meta JSON files")
    parser.add_argument(
        "system_config",
        type=valid_system_config,
        help=(
            'Path to the system config JSON file with the specified software '
            'zone name. Format as follows: "<path>@<zone_name>", e.g., '
            '"/path/to/system_config.json@Z01".'
            )
        )
    parser.add_argument(
        "--visu-types",
        type=str,
        required=False,
        default=None,
        help=(
            'Path to a local `visu-types-g3.svg` file. If not provided, '
            'the file will be downloaded from the Gitlab repository.'
            )
        )
    parser.add_argument(
        "-f",
        "--file",
        type=str,
        required=False,
        choices=['meta', 'typeinfo', 'site-svg'],
        help=(
            'Specify the type of file to generate. If not provided, '
            'all files will be generated.'
            )
        )
    parser.add_argument(
        "-o",
        "--output-dir",
        type=str,
        default=os.getcwd(),
        help="Path to the output directory",
        )
    parser.add_argument(
        "-u",
        "--user",
        type=str,
        required=False,
        help="Siteware default SHV user name",
        )
    parser.add_argument(
        "-p",
        "--password",
        type=str,
        required=False,
        help="Siteware default SHV user password",
        )
    parser.add_argument(
        '--log-level',
        type=str,
        required=False,
        default='INFO',
        choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
        help='Set the logging level.'
        )
    return parser.parse_args()


def main() -> None:
    # parse and validate the system config
    args = parse_args()
    logging.basicConfig(
        level=args.log_level,
        format='[%(name)s] %(levelname)s:%(message)s'
        )
    system_config_path, zone_name = args.system_config
    if not os.path.isfile(system_config_path):
        logger.error(f"System config file not found: {system_config_path}")
        raise SystemExit(1)
    with open(system_config_path, 'r', encoding='utf-8') as f:
        system_config = json.load(f)
    ensure_token_set()
    # create output directories and file paths
    zone_dir = os.path.join(args.output_dir, zone_name)
    zone_file_dir = os.path.join(zone_dir, '_files')
    os.makedirs(zone_file_dir, exist_ok=True)
    # create and write the files
    if args.file is None or args.file == 'meta':
        create_meta_files(system_config, zone_name, args.output_dir, args.user, args.password)
    if args.file is None or args.file == 'typeinfo':
        create_type_info_file(system_config, zone_name, zone_file_dir)
    if args.file is None or args.file == 'site-svg':
        create_site_svg_file(
            system_config, zone_name, zone_file_dir, args.visu_types
            )



import enum
import gitlab
import logging
import os
import typing

from platformdirs import PlatformDirs

if typing.TYPE_CHECKING:
    from gitlab.v4.objects import Project

logger = logging.getLogger('typeinfo_parser.utils')


Dirs = PlatformDirs("SystemG3Visu", ensure_exists=True)


class AccessToken:

    @staticmethod
    def get() -> str:
        path = os.path.join(Dirs.user_data_dir, ".token")
        if not os.path.isfile(path):
            logger.warning('Token file was not found at "%s".', path)
            return ''
        with open(path, "r", encoding='utf-8') as file:
            return file.read().strip()

    @staticmethod
    def set(token: str) -> None:
        path = os.path.join(Dirs.user_data_dir, ".token")
        with open(path, "w", encoding='utf-8') as file:
            file.write(token)

    @staticmethod
    def is_valid(token: str) -> bool:
        try:
            with gitlab.Gitlab(SitesFileDownloader.BASE_URL, token) as gl:
                gl.projects.get(SitesFileDownloader.PROJECT_ID)
                return True
        except gitlab.GitlabAuthenticationError:
            return False
        except gitlab.GitlabGetError:  # project may be invisible
            return False               # without the correct access token


def prompt_token_until_valid() -> str:
    while True:
        try:
            token = input("Invalid token. Enter your Gitlab access token: ")
            if AccessToken.is_valid(token):
                break
        except KeyboardInterrupt:
            raise SystemExit(1)
        except Exception as exc:
            logger.error('Unexpected error occured: {%s}', exc)
    return token


def ensure_token_set() -> str:
    token = AccessToken.get()
    if not token or not AccessToken.is_valid(token):
        token = prompt_token_until_valid()
        AccessToken.set(token)
    return token


class SitesFilePaths(enum.StrEnum):
    """Relative paths to the files in the repository."""
    TYPE_INFO = "sites/_files/typeInfo.cpon"
    """Path to the `visu-types-g3.svg` file in the repository."""
    VISU_TYPES_G3 = "sites/_files/predefined-visu-types/visu-types-g3.svg"
    """Path to the `visu-types-g3.svg` file in the repository."""


class SitesFileDownloader:
    BASE_URL: str = "https://gitlab.com"
    """Base URL where the repository is hosted."""
    PROJECT_ID: int = 9492559
    """ID of the project repository on Gitlab."""

    @staticmethod
    def _get_contents(
        project: 'Project', file_path: str, branch: str, decode: bool
    ) -> str | bytes | None:
        """
        Retrieve the contents of a file from a project.

        Args:
            project (Project): Project object.
            file_path (str): Path to the file in the project repository.
            branch (str): Branch name.

        Returns:
            bytes: The contents of the file.
        """
        try:
            logger.info(f'Retrieving file from "{file_path}".')
            contents = project.files.raw(file_path=file_path, ref=branch)
            if not isinstance(contents, bytes):
                raise TypeError(
                    f'Unexpected file data type: "{type(contents).__name__}" '
                    f'(expected type "bytes")'
                    )
            if decode:
                return contents.decode()
            return contents
        except gitlab.GitlabGetError:
            logger.error('Failed to retrieve file data from "%s".', file_path)
        except gitlab.GitlabAuthenticationError:
            project_name = project.asdict().get("name", "UNKNOWN_PROJECT")
            logger.error('Authentication error to project "%s".', project_name)
        except Exception as exc:
            logger.error(
                'Unexpected error occured while retrieving file contents: '
                '{%s}', exc
                )
        return None

    @staticmethod
    def _extract_file_name(path: str) -> str:
        _, file_name = os.path.split(path)
        return file_name

    @staticmethod
    def _check_save_to_path(
        file_name: str, save_to: str, overwrite_if_exists: bool
    ) -> str:
        logger.info('Ensuring local path "%s" exists.', save_to)
        os.makedirs(save_to, exist_ok=True)
        file_path = os.path.join(save_to, file_name)
        if os.path.isfile(file_path):
            msg = f'File "{file_name}" already exists in "{save_to}"'
            if overwrite_if_exists:
                msg += ', but will be overwritten.'
                logger.warning(msg)
            else:
                msg += '.'
                raise FileExistsError(msg)
        return file_path

    def read(
        self,
        file_path: str,
        branch: str = 'master',
        access_token: str | None = None,
        decode: bool = True
    ) -> str | bytes | None:
        """Read the contents of a file from the repository.

        Args:
            file_path (str): Path to the file in the repository.
            branch (str, optional): Reference Branch name.
                Defaults to 'master'.
            access_token (str | None, optional): Gitlab access token.
            decode (bool, optional): Whether to decode the file contents from
                bytes to string. Defaults to True.

        Returns:
            str | bytes | None: File contents. If the file was not found or
                an error occured, None is returned.
        """
        access_token = access_token or AccessToken.get()
        logger.info(f'Initializing a Gitlab instance for "{self.BASE_URL}"')
        with gitlab.Gitlab(self.BASE_URL, access_token) as gl:
            logger.info(f"Retrieving the project with id={self.PROJECT_ID}")
            project = gl.projects.get(self.PROJECT_ID)
            contents = self._get_contents(project, file_path, branch, decode)
        return contents

    def download(
        self,
        file_path: str,
        branch: str = 'master',
        save_to: str = os.getcwd(),
        overwrite_if_exists: bool = False
    ) -> None:
        """Download a file from the repository.

        Args:
            file_path (str): Path to the file in the repository.
            branch (str, optional): Reference Branch name.
                Defaults to 'master'.
            save_to (str, optional): Local path to save the file to.
                Defaults to the current working directory.
            overwrite_if_exists (bool, optional): Whether to overwrite the file
                if it already exists in the save_to directory.
                Defaults to False.
        """
        file_name = self._extract_file_name(file_path)
        if not file_name:
            logger.error('File name cannot be empty.')
            return
        try:
            local_path = self._check_save_to_path(
                file_name, save_to, overwrite_if_exists
                )
        except FileExistsError as err:
            logger.error(err)
            return
        contents = self.read(file_path, branch, decode=False)
        if contents is None:
            logger.error('Failed to save file contents to "%s".', local_path)
            return
        try:
            logger.info('Saving file contents to "%s".', local_path)
            assert isinstance(contents, bytes)
            with open(local_path, 'wb') as file:
                file.write(contents)
        except Exception as exc:
            logger.error(
                'Failed to save file contents to "%s" (%s).', local_path, exc
                )
